import dataclasses
from typing import Any, Callable, Dict, List, Optional, Union

from .. import datetime_helpers, llm_usage
from ..api_objects import helpers, span, attachment
from ..types import ErrorInfoDict, SpanType, DistributedTraceHeadersDict


@dataclasses.dataclass
class BaseArguments:
    def to_kwargs(self, ignore_keys: Optional[List[str]] = None) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        ignore_keys = [] if ignore_keys is None else ignore_keys
        for key, value in self.__dict__.items():
            if (value is not None) and (key not in ignore_keys):
                result[key] = value

        return result


@dataclasses.dataclass
class EndSpanParameters(BaseArguments):
    """
    Span data parameters that we set (or might set) when the
    tracked function is ended.
    """

    name: Optional[str] = None
    metadata: Optional[Any] = None
    input: Optional[Dict[str, Any]] = None
    output: Optional[Dict[str, Any]] = None
    tags: Optional[List[str]] = None
    usage: Optional[Union[Dict[str, Any], llm_usage.OpikUsage]] = None
    model: Optional[str] = None
    provider: Optional[str] = None
    error_info: Optional[ErrorInfoDict] = None
    total_cost: Optional[float] = None
    attachments: Optional[List[attachment.Attachment]] = None


@dataclasses.dataclass
class StartSpanParameters(BaseArguments):
    """
    Span data parameters that we set (or might set) when the
    tracked function is started.
    """

    type: SpanType
    name: str
    tags: Optional[List[str]] = None
    metadata: Optional[Dict[str, Any]] = None
    input: Optional[Dict[str, Any]] = None
    project_name: Optional[str] = None
    model: Optional[str] = None
    provider: Optional[str] = None
    thread_id: Optional[str] = None  # used for traces only


@dataclasses.dataclass
class TrackOptions(BaseArguments):
    """
    A storage for all arguments passed to the `track` decorator.
    """

    name: Optional[str]
    type: SpanType
    tags: Optional[List[str]]
    metadata: Optional[Dict[str, Any]]
    capture_input: bool
    ignore_arguments: Optional[List[str]]
    capture_output: bool
    generations_aggregator: Optional[Callable[[List[Any]], Any]]
    flush: bool
    project_name: Optional[str]
    create_duplicate_root_span: bool


def create_span_data(
    start_span_arguments: StartSpanParameters,
    trace_id: str,
    parent_span_id: Optional[str] = None,
) -> span.SpanData:
    span_data = span.SpanData(
        id=helpers.generate_id(),
        parent_span_id=parent_span_id,
        trace_id=trace_id,
        start_time=datetime_helpers.local_timestamp(),
        name=start_span_arguments.name,
        type=start_span_arguments.type,
        tags=start_span_arguments.tags,
        metadata=start_span_arguments.metadata,
        input=start_span_arguments.input,
        project_name=start_span_arguments.project_name,
        model=start_span_arguments.model,
        provider=start_span_arguments.provider,
    )
    return span_data


def extract_distributed_trace_headers(
    kwargs: Dict[str, Any],
) -> Optional[DistributedTraceHeadersDict]:
    return kwargs.pop("opik_distributed_trace_headers", None)



from typing import Dict, List, Literal, Optional

import pydantic

ConversationDict = Dict[Literal["role", "content"], str]
Conversation = List[ConversationDict]


class ConversationTurn(pydantic.BaseModel):
    """
    Representation of a single turn in a conversation.
    This class defines a model for encapsulating a single conversational
    turn consisting of an input user's message and an output LLM message. It is
    designed to handle the exchange of messages in a structured format.
    Args:
        input: The input message of the conversation turn.
        output: The output message of the conversation turn.
    Example:
        >>> conversation_turn = ConversationTurn(
        >>>     input={"role": "user", "content": "Hello!"},
        >>>     output={"role": "assistant", "content": "Hi there! How can I help you today?"}
        >>> )
    """

    input: ConversationDict
    output: Optional[ConversationDict]

    def as_list(self) -> List[ConversationDict]:
        if self.output is None:
            return [self.input]
        return [self.input, self.output]


__all__ = ["ConversationDict", "Conversation", "ConversationTurn"]



import abc
import logging
from contextlib import contextmanager, asynccontextmanager
from typing import Any, List, Dict, Optional, Type
import pydantic

from opik import exceptions


LOGGER = logging.getLogger(__name__)


class OpikBaseModel(abc.ABC):
    """
    This class serves as an interface to LLMs.

    If you want to implement a custom LLM provider in evaluation metrics,
    you should inherit from this class.
    """

    def __init__(self, model_name: str):
        """
        Initializes the base model with a given model name.

        Args:
            model_name: The name of the LLM to be used.
        """
        self.model_name = model_name

    @abc.abstractmethod
    def generate_string(
        self,
        input: str,
        response_format: Optional[Type[pydantic.BaseModel]] = None,
        **kwargs: Any,
    ) -> str:
        """
        Simplified interface to generate a string output from the model.

        Args:
            input: The input string based on which the model will generate the output.
            kwargs: Additional arguments that may be used by the model for string generation.

        Returns:
            str: The generated string output.
        """
        pass

    @abc.abstractmethod
    def generate_provider_response(
        self, messages: List[Dict[str, Any]], **kwargs: Any
    ) -> Any:
        """
        Do not use this method directly. It is intended to be used within `get_provider_response()` method.

        Generate a provider-specific response. Can be used to interface with
        the underlying model provider (e.g., OpenAI, Anthropic) and get raw output.

        Args:
            messages: A list of messages to be sent to the model, should be a list of dictionaries with the keys
            kwargs: arguments required by the provider to generate a response.

        Returns:
            Any: The response from the model provider, which can be of any type depending on the use case and LLM.
        """
        pass

    async def agenerate_string(
        self,
        input: str,
        response_format: Optional[Type[pydantic.BaseModel]] = None,
        **kwargs: Any,
    ) -> str:
        """
        Simplified interface to generate a string output from the model. Async version.

        Args:
            input: The input string based on which the model will generate the output.
            kwargs: Additional arguments that may be used by the model for string generation.

        Returns:
            str: The generated string output.
        """
        raise NotImplementedError("Async generation not implemented for this provider")

    async def agenerate_provider_response(
        self, messages: List[Dict[str, Any]], **kwargs: Any
    ) -> Any:
        """
        Do not use this method directly. It is intended to be used within `aget_provider_response()` method.

        Generate a provider-specific response. Can be used to interface with
        the underlying model provider (e.g., OpenAI, Anthropic) and get raw output.
        Async version.

        Args:
            messages: A list of messages to be sent to the model, should be a list of dictionaries with the keys
                "content" and "role".
            kwargs: arguments required by the provider to generate a response.

        Returns:
            Any: The response from the model provider, which can be of any type depending on the use case and LLM.
        """
        raise NotImplementedError("Async generation not implemented for this provider")


@contextmanager
def get_provider_response(
    model_provider: OpikBaseModel, messages: List[Dict[str, Any]], **kwargs: Any
) -> Any:
    """
    Provides a context manager for getting and managing the response from a
    model provider. Ensures that errors during the interaction with the model
    provider are handled appropriately and logged.

    Args:
        model_provider: Instance of a class derived from `OpikBaseModel`
            responsible for interfacing with the model.
        messages: List of dictionaries containing the messages or inputs to be
            passed to the model.
        **kwargs: Additional keyword arguments to customize the generation of
            the model responses.

    Yields:
        Any: The response generated by the model provider.

    Raises:
        exceptions.BaseLLMError: If the response generation from the model provider
            fails due to an exception.
    """
    try:
        yield model_provider.generate_provider_response(messages, **kwargs)
    except Exception as e:
        LOGGER.error("Failed to call LLM provider, reason: %s", e)
        raise exceptions.BaseLLMError(str(e))


@asynccontextmanager
async def aget_provider_response(
    model_provider: OpikBaseModel, messages: List[Dict[str, Any]], **kwargs: Any
) -> Any:
    """
    Asynchronous context manager for getting a response from a model provider.

    This function asynchronously interacts with the specified `model_provider` to
    generate a response based on the given list of `messages` and additional
    optional keyword arguments. If an error occurs during this process, it is
    logged, and a custom exception is raised.

    Args:
        model_provider: The model provider from which to request
            the response.
        messages: A list of dictionaries containing the
            messages for the model provider to process.
        **kwargs: Additional keyword arguments passed to the model provider's
            response generation method.

    Yields:
        Any: The response generated asynchronously by the model provider.

    Raises:
        exceptions.BaseLLMError: If there is an error during the asynchronous
            interaction with the model provider.
    """
    try:
        response = await model_provider.agenerate_provider_response(
            messages=messages, **kwargs
        )
        yield response
    except Exception as e:
        LOGGER.error("Failed to call LLM provider asynchronously, reason: %s", e)
        raise exceptions.BaseLLMError(str(e))


def check_model_output_string(output: Optional[str]) -> str:
    """
    Checks the output of a model and verifies that it is not None.

    This function ensures that the output returned from a language model (LLM) has a valid, non-null value.
    If the output is found to be None, an error is raised with a detailed message. This can help in
    debugging issues related to incorrect environment configuration or missing API keys.

    Args:
        output: The output string generated by the language model to be validated.

    Returns:
        The output of the language model that was validated.

    Raises:
        exceptions.BaseLLMError: Raised if the output is evaluated to None. The error message contains suggestions to
        verify environment configurations and check model API key availability.
    """
    if output is None:
        raise exceptions.BaseLLMError(
            "Received None as the output from the LLM. Please verify your environment configuration "
            "and ensure that the API keys for the models in use (e.g., OPENAI_API_KEY) are set correctly."
        )

    return output



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..errors.bad_request_error import BadRequestError
from ..errors.conflict_error import ConflictError
from ..errors.not_found_error import NotFoundError
from ..errors.unprocessable_entity_error import UnprocessableEntityError
from ..types.breakdown_config_public import BreakdownConfigPublic
from ..types.feedback_score_names import FeedbackScoreNames
from ..types.project_detailed import ProjectDetailed
from ..types.project_metric_response_public import ProjectMetricResponsePublic
from ..types.project_page_public import ProjectPagePublic
from ..types.project_public import ProjectPublic
from ..types.project_stats_summary import ProjectStatsSummary
from ..types.span_filter_public import SpanFilterPublic
from ..types.token_usage_names import TokenUsageNames
from ..types.trace_filter_public import TraceFilterPublic
from ..types.trace_thread_filter_public import TraceThreadFilterPublic
from .types.project_metric_request_public_interval import ProjectMetricRequestPublicInterval
from .types.project_metric_request_public_metric_type import ProjectMetricRequestPublicMetricType
from .types.project_update_visibility import ProjectUpdateVisibility
from .types.project_write_visibility import ProjectWriteVisibility

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawProjectsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def find_projects(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[ProjectPagePublic]:
        """
        Find projects

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProjectPagePublic]
            Project resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/projects",
            method="GET",
            params={
                "page": page,
                "size": size,
                "name": name,
                "sorting": sorting,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectPagePublic,
                    parse_obj_as(
                        type_=ProjectPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_project(
        self,
        *,
        name: str,
        visibility: typing.Optional[ProjectWriteVisibility] = OMIT,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Create project

        Parameters
        ----------
        name : str

        visibility : typing.Optional[ProjectWriteVisibility]

        description : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/projects",
            method="POST",
            json={
                "name": name,
                "visibility": visibility,
                "description": description,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_project_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[ProjectPublic]:
        """
        Get project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProjectPublic]
            Project resource
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectPublic,
                    parse_obj_as(
                        type_=ProjectPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_project_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Delete project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def update_project(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        visibility: typing.Optional[ProjectUpdateVisibility] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Update project by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        description : typing.Optional[str]

        visibility : typing.Optional[ProjectUpdateVisibility]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}",
            method="PATCH",
            json={
                "name": name,
                "description": description,
                "visibility": visibility,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def delete_projects_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Delete projects batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/projects/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def find_feedback_score_names_by_project_ids(
        self, *, project_ids: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[FeedbackScoreNames]:
        """
        Find Feedback Score names By Project Ids

        Parameters
        ----------
        project_ids : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[FeedbackScoreNames]
            Feedback Scores resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/projects/feedback-scores/names",
            method="GET",
            params={
                "project_ids": project_ids,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FeedbackScoreNames,
                    parse_obj_as(
                        type_=FeedbackScoreNames,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def find_token_usage_names(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[TokenUsageNames]:
        """
        Find Token Usage names

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TokenUsageNames]
            Token Usage names resource
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}/token-usage/names",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TokenUsageNames,
                    parse_obj_as(
                        type_=TokenUsageNames,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_project_metrics(
        self,
        id: str,
        *,
        metric_type: typing.Optional[ProjectMetricRequestPublicMetricType] = OMIT,
        interval: typing.Optional[ProjectMetricRequestPublicInterval] = OMIT,
        interval_start: typing.Optional[dt.datetime] = OMIT,
        interval_end: typing.Optional[dt.datetime] = OMIT,
        span_filters: typing.Optional[typing.Sequence[SpanFilterPublic]] = OMIT,
        trace_filters: typing.Optional[typing.Sequence[TraceFilterPublic]] = OMIT,
        thread_filters: typing.Optional[typing.Sequence[TraceThreadFilterPublic]] = OMIT,
        breakdown: typing.Optional[BreakdownConfigPublic] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[ProjectMetricResponsePublic]:
        """
        Gets specified metrics for a project

        Parameters
        ----------
        id : str

        metric_type : typing.Optional[ProjectMetricRequestPublicMetricType]

        interval : typing.Optional[ProjectMetricRequestPublicInterval]

        interval_start : typing.Optional[dt.datetime]

        interval_end : typing.Optional[dt.datetime]

        span_filters : typing.Optional[typing.Sequence[SpanFilterPublic]]

        trace_filters : typing.Optional[typing.Sequence[TraceFilterPublic]]

        thread_filters : typing.Optional[typing.Sequence[TraceThreadFilterPublic]]

        breakdown : typing.Optional[BreakdownConfigPublic]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProjectMetricResponsePublic]
            Project Metrics
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}/metrics",
            method="POST",
            json={
                "metric_type": metric_type,
                "interval": interval,
                "interval_start": interval_start,
                "interval_end": interval_end,
                "span_filters": convert_and_respect_annotation_metadata(
                    object_=span_filters, annotation=typing.Sequence[SpanFilterPublic], direction="write"
                ),
                "trace_filters": convert_and_respect_annotation_metadata(
                    object_=trace_filters, annotation=typing.Sequence[TraceFilterPublic], direction="write"
                ),
                "thread_filters": convert_and_respect_annotation_metadata(
                    object_=thread_filters, annotation=typing.Sequence[TraceThreadFilterPublic], direction="write"
                ),
                "breakdown": convert_and_respect_annotation_metadata(
                    object_=breakdown, annotation=BreakdownConfigPublic, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectMetricResponsePublic,
                    parse_obj_as(
                        type_=ProjectMetricResponsePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_project_stats(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[ProjectStatsSummary]:
        """
        Get Project Stats

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProjectStatsSummary]
            Project Stats
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/projects/stats",
            method="GET",
            params={
                "page": page,
                "size": size,
                "name": name,
                "sorting": sorting,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectStatsSummary,
                    parse_obj_as(
                        type_=ProjectStatsSummary,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def retrieve_project(
        self, *, name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[ProjectDetailed]:
        """
        Retrieve project

        Parameters
        ----------
        name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[ProjectDetailed]
            Project resource
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/projects/retrieve",
            method="POST",
            json={
                "name": name,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectDetailed,
                    parse_obj_as(
                        type_=ProjectDetailed,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawProjectsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def find_projects(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[ProjectPagePublic]:
        """
        Find projects

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProjectPagePublic]
            Project resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/projects",
            method="GET",
            params={
                "page": page,
                "size": size,
                "name": name,
                "sorting": sorting,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectPagePublic,
                    parse_obj_as(
                        type_=ProjectPagePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_project(
        self,
        *,
        name: str,
        visibility: typing.Optional[ProjectWriteVisibility] = OMIT,
        description: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Create project

        Parameters
        ----------
        name : str

        visibility : typing.Optional[ProjectWriteVisibility]

        description : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/projects",
            method="POST",
            json={
                "name": name,
                "visibility": visibility,
                "description": description,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_project_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[ProjectPublic]:
        """
        Get project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProjectPublic]
            Project resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectPublic,
                    parse_obj_as(
                        type_=ProjectPublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_project_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Delete project by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}",
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def update_project(
        self,
        id: str,
        *,
        name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        visibility: typing.Optional[ProjectUpdateVisibility] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Update project by id

        Parameters
        ----------
        id : str

        name : typing.Optional[str]

        description : typing.Optional[str]

        visibility : typing.Optional[ProjectUpdateVisibility]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}",
            method="PATCH",
            json={
                "name": name,
                "description": description,
                "visibility": visibility,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def delete_projects_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Delete projects batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/projects/delete",
            method="POST",
            json={
                "ids": ids,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def find_feedback_score_names_by_project_ids(
        self, *, project_ids: typing.Optional[str] = None, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[FeedbackScoreNames]:
        """
        Find Feedback Score names By Project Ids

        Parameters
        ----------
        project_ids : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[FeedbackScoreNames]
            Feedback Scores resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/projects/feedback-scores/names",
            method="GET",
            params={
                "project_ids": project_ids,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    FeedbackScoreNames,
                    parse_obj_as(
                        type_=FeedbackScoreNames,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def find_token_usage_names(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[TokenUsageNames]:
        """
        Find Token Usage names

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TokenUsageNames]
            Token Usage names resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}/token-usage/names",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TokenUsageNames,
                    parse_obj_as(
                        type_=TokenUsageNames,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_project_metrics(
        self,
        id: str,
        *,
        metric_type: typing.Optional[ProjectMetricRequestPublicMetricType] = OMIT,
        interval: typing.Optional[ProjectMetricRequestPublicInterval] = OMIT,
        interval_start: typing.Optional[dt.datetime] = OMIT,
        interval_end: typing.Optional[dt.datetime] = OMIT,
        span_filters: typing.Optional[typing.Sequence[SpanFilterPublic]] = OMIT,
        trace_filters: typing.Optional[typing.Sequence[TraceFilterPublic]] = OMIT,
        thread_filters: typing.Optional[typing.Sequence[TraceThreadFilterPublic]] = OMIT,
        breakdown: typing.Optional[BreakdownConfigPublic] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[ProjectMetricResponsePublic]:
        """
        Gets specified metrics for a project

        Parameters
        ----------
        id : str

        metric_type : typing.Optional[ProjectMetricRequestPublicMetricType]

        interval : typing.Optional[ProjectMetricRequestPublicInterval]

        interval_start : typing.Optional[dt.datetime]

        interval_end : typing.Optional[dt.datetime]

        span_filters : typing.Optional[typing.Sequence[SpanFilterPublic]]

        trace_filters : typing.Optional[typing.Sequence[TraceFilterPublic]]

        thread_filters : typing.Optional[typing.Sequence[TraceThreadFilterPublic]]

        breakdown : typing.Optional[BreakdownConfigPublic]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProjectMetricResponsePublic]
            Project Metrics
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/projects/{jsonable_encoder(id)}/metrics",
            method="POST",
            json={
                "metric_type": metric_type,
                "interval": interval,
                "interval_start": interval_start,
                "interval_end": interval_end,
                "span_filters": convert_and_respect_annotation_metadata(
                    object_=span_filters, annotation=typing.Sequence[SpanFilterPublic], direction="write"
                ),
                "trace_filters": convert_and_respect_annotation_metadata(
                    object_=trace_filters, annotation=typing.Sequence[TraceFilterPublic], direction="write"
                ),
                "thread_filters": convert_and_respect_annotation_metadata(
                    object_=thread_filters, annotation=typing.Sequence[TraceThreadFilterPublic], direction="write"
                ),
                "breakdown": convert_and_respect_annotation_metadata(
                    object_=breakdown, annotation=BreakdownConfigPublic, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectMetricResponsePublic,
                    parse_obj_as(
                        type_=ProjectMetricResponsePublic,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_project_stats(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[ProjectStatsSummary]:
        """
        Get Project Stats

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        sorting : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProjectStatsSummary]
            Project Stats
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/projects/stats",
            method="GET",
            params={
                "page": page,
                "size": size,
                "name": name,
                "sorting": sorting,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectStatsSummary,
                    parse_obj_as(
                        type_=ProjectStatsSummary,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def retrieve_project(
        self, *, name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[ProjectDetailed]:
        """
        Retrieve project

        Parameters
        ----------
        name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[ProjectDetailed]
            Project resource
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/projects/retrieve",
            method="POST",
            json={
                "name": name,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    ProjectDetailed,
                    parse_obj_as(
                        type_=ProjectDetailed,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 422:
                raise UnprocessableEntityError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



import logging
from typing import Any, Dict, List, TypeVar

from langchain_core.runnables import base as runnables_base

from . import opik_tracer as opik_tracer_module

LOGGER = logging.getLogger(__name__)

CompiledGraphType = TypeVar("CompiledGraphType", bound=runnables_base.Runnable)


def track_langgraph(
    graph: CompiledGraphType,
    opik_tracer: opik_tracer_module.OpikTracer,
) -> CompiledGraphType:
    """
    Adds Opik tracking to a compiled LangGraph graph by injecting OpikTracer into its default config.

    After calling this function, all subsequent invocations of the graph will automatically
    be tracked without needing to pass the OpikTracer in the config parameter.

    The function will automatically extract the graph structure visualization from the compiled
    graph if it wasn't already provided when creating the OpikTracer. This visualization will
    be included in the trace metadata in the Opik UI.

    Args:
        graph: A compiled LangGraph graph (result of StateGraph.compile()).
        opik_tracer: An OpikTracer instance to use for tracking the graph.

    Returns:
        The modified graph with Opik tracking enabled.

    Example:
        ```python
        from langgraph.graph import StateGraph, START, END
        from opik.integrations.langchain import OpikTracer, track_langgraph

        # Build your graph
        builder = StateGraph(State)
        builder.add_node("my_node", my_node_function)
        builder.add_edge(START, "my_node")
        builder.add_edge("my_node", END)

        # Compile the graph
        graph = builder.compile()

        # Create OpikTracer and track the graph once
        # No need to manually extract the graph - it's done automatically!
        opik_tracer = OpikTracer(
            tags=["production"],
            metadata={"version": "1.0"}
        )
        graph = track_langgraph(graph, opik_tracer)

        # Now all invocations are tracked automatically
        result = graph.invoke({"message": "Hello"})
        # No need to pass config={"callbacks": [opik_tracer]} anymore!
        ```

    Note:
        - The graph visualization is automatically extracted and added to trace metadata
          if not already provided in the OpikTracer constructor.
        - If you need to customize the OpikTracer for specific invocations, you can still
          pass it explicitly in the config parameter, which will override the default.
        - The graph object is modified in-place and also returned for convenience.
        - For async invocations using `ainvoke()`, you may still need to use
          `extract_current_langgraph_span_data()` to propagate context to @track-decorated
          functions within async nodes.
    """
    graph_structure = graph.get_graph(xray=True)
    opik_tracer.set_graph(graph_structure)

    # Inject the callback into the graph's default config
    config: Dict[str, Any] = getattr(graph, "config", None) or {}
    graph.config = config  # type: ignore[attr-defined]
    callbacks: List[Any] = config.setdefault("callbacks", [])

    if any(isinstance(cb, opik_tracer_module.OpikTracer) for cb in callbacks):
        LOGGER.warning(
            "Graph already has an OpikTracer callback injected. "
            "Skipping re-tracking to avoid duplicate callbacks."
        )
        return graph

    callbacks.append(opik_tracer)

    return graph



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .experiment_write_status import ExperimentWriteStatus
from .experiment_write_type import ExperimentWriteType

__all__ = ["ExperimentWriteStatus", "ExperimentWriteType"]



"""Prompt export functionality."""

import json
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, List, Optional, Union

import click
from rich.console import Console

import opik
from opik.api_objects.prompt import Prompt, ChatPrompt
from .utils import (
    debug_print,
    prompt_to_csv_rows,
    should_skip_file,
    write_csv_data,
    write_json_data,
    print_export_summary,
)

console = Console()


def _get_prompt_content(prompt: Any) -> Any:
    """Extract prompt content based on prompt type.

    Args:
        prompt: A Prompt or ChatPrompt instance

    Returns:
        For Prompt: the prompt string
        For ChatPrompt: the template (list of message dicts)
        Otherwise: None
    """
    if isinstance(prompt, Prompt):
        return prompt.prompt
    elif isinstance(prompt, ChatPrompt):
        return prompt.template
    return None


def _get_template_structure(prompt: Any) -> str:
    """Get template_structure based on prompt type.

    Args:
        prompt: A Prompt or ChatPrompt instance

    Returns:
        "text" for Prompt, "chat" for ChatPrompt, "text" as default
    """
    if isinstance(prompt, ChatPrompt):
        return "chat"
    elif isinstance(prompt, Prompt):
        return "text"
    return "text"


def _get_prompt_type_string(prompt: Any) -> Optional[str]:
    """Get prompt type as uppercase string.

    Args:
        prompt: A Prompt or ChatPrompt instance

    Returns:
        Uppercase type string (e.g., "JINJA2", "MUSTACHE") or None
    """
    prompt_type = getattr(prompt, "type", None)
    if prompt_type is None:
        return None

    # If it's an enum, get the value and convert to uppercase
    if hasattr(prompt_type, "value"):
        return prompt_type.value.upper()
    # If it's already a string, convert to uppercase
    if isinstance(prompt_type, str):
        return prompt_type.upper()
    # Otherwise, convert to string and uppercase
    return str(prompt_type).upper()


def export_single_prompt(
    client: opik.Opik,
    prompt: Union[Prompt, ChatPrompt],
    output_dir: Path,
    max_results: Optional[int],
    force: bool,
    debug: bool,
    format: str,
) -> int:
    """Export a single prompt."""
    try:
        # Check if already exists and force is not set
        if format.lower() == "csv":
            prompt_file = output_dir / f"prompts_{prompt.name.replace('/', '_')}.csv"
        else:
            prompt_file = output_dir / f"prompt_{prompt.name.replace('/', '_')}.json"

        if should_skip_file(prompt_file, force):
            if debug:
                debug_print(f"Skipping {prompt.name} (already exists)", debug)
            return 0

        # Get prompt history - use appropriate method based on prompt type
        prompt_history: List[Union[Prompt, ChatPrompt]]
        if isinstance(prompt, ChatPrompt):
            prompt_history = list(client.get_chat_prompt_history(prompt.name))
        else:
            prompt_history = list(client.get_prompt_history(prompt.name))

        # Create prompt data structure
        prompt_data = {
            "name": prompt.name,
            "current_version": {
                "prompt": _get_prompt_content(prompt),
                "metadata": getattr(prompt, "metadata", None),
                "type": _get_prompt_type_string(prompt),
                "commit": getattr(prompt, "commit", None),
                "template_structure": _get_template_structure(prompt),
            },
            "history": [
                {
                    "prompt": _get_prompt_content(version),
                    "metadata": getattr(version, "metadata", None),
                    "type": _get_prompt_type_string(version),
                    "commit": getattr(version, "commit", None),
                    "template_structure": _get_template_structure(version),
                }
                for version in prompt_history
            ],
            "downloaded_at": datetime.now().isoformat(),
        }

        # Save to file using the appropriate format
        if format.lower() == "csv":
            write_csv_data(prompt_data, prompt_file, prompt_to_csv_rows)
        else:
            write_json_data(prompt_data, prompt_file)

        if debug:
            debug_print(f"Exported prompt: {prompt.name}", debug)
        return 1

    except Exception as e:
        console.print(f"[red]Error exporting prompt {prompt.name}: {e}[/red]")
        return 0


def export_prompt_by_name(
    name: str,
    workspace: str,
    output_path: str,
    max_results: Optional[int],
    force: bool,
    debug: bool,
    format: str,
    api_key: Optional[str] = None,
) -> None:
    """Export a prompt by exact name."""
    try:
        if debug:
            debug_print(f"Exporting prompt: {name}", debug)

        # Initialize client
        if api_key:
            client = opik.Opik(api_key=api_key, workspace=workspace)
        else:
            client = opik.Opik(workspace=workspace)

        # Create output directory
        output_dir = Path(output_path) / workspace / "prompts"
        output_dir.mkdir(parents=True, exist_ok=True)

        if debug:
            debug_print(f"Target directory: {output_dir}", debug)

        # Try to get prompt by exact name
        # Try ChatPrompt first, then regular Prompt
        prompt: Optional[Union[Prompt, ChatPrompt]] = None
        try:
            prompt = client.get_chat_prompt(name)
            if debug and prompt:
                debug_print(f"Found ChatPrompt by direct lookup: {prompt.name}", debug)
        except Exception:
            # Not a ChatPrompt, try regular Prompt
            try:
                prompt = client.get_prompt(name)
                if not prompt:
                    console.print(f"[red]Prompt '{name}' not found[/red]")
                    return
                if debug:
                    debug_print(f"Found Prompt by direct lookup: {prompt.name}", debug)
            except Exception as e:
                console.print(f"[red]Prompt '{name}' not found: {e}[/red]")
                return

        if prompt is None:
            console.print(f"[red]Prompt '{name}' not found[/red]")
            return

        # Export the prompt
        exported_count = export_single_prompt(
            client, prompt, output_dir, max_results, force, debug, format
        )

        # Collect statistics for summary
        stats = {
            "prompts": 1 if exported_count > 0 else 0,
            "prompts_skipped": 0 if exported_count > 0 else 1,
        }

        # Show export summary
        print_export_summary(stats, format)

        if exported_count > 0:
            console.print(
                f"[green]Successfully exported prompt '{name}' to {output_dir}[/green]"
            )
        else:
            console.print(
                f"[yellow]Prompt '{name}' already exists (use --force to re-download)[/yellow]"
            )

    except Exception as e:
        console.print(f"[red]Error exporting prompt: {e}[/red]")
        sys.exit(1)


def export_prompts_by_ids(
    client: opik.Opik,
    prompt_ids: set[str],
    prompts_dir: Path,
    format: str,
    debug: bool,
    force: bool,
) -> tuple[int, int]:
    """Export prompts by their IDs.

    Args:
        client: Opik client instance
        prompt_ids: Set of prompt IDs to export
        prompts_dir: Directory to save prompts
        format: Export format ('json' or 'csv')
        debug: Enable debug output
        force: Re-download prompts even if they already exist locally

    Returns:
        Tuple of (exported_count, skipped_count)
    """
    exported_count = 0
    skipped_count = 0

    for prompt_id in prompt_ids:
        try:
            # Get the prompt - try ChatPrompt first, then regular Prompt
            prompt: Optional[Union[Prompt, ChatPrompt]] = None
            try:
                prompt = client.get_chat_prompt(prompt_id)
            except Exception:
                # Not a ChatPrompt, try regular Prompt
                prompt = client.get_prompt(prompt_id)

            if not prompt:
                if debug:
                    console.print(
                        f"[yellow]Warning: Prompt {prompt_id} not found[/yellow]"
                    )
                continue

            # Determine file path
            if format.lower() == "csv":
                prompt_file = (
                    prompts_dir
                    / f"prompts_{prompt.name or getattr(prompt, 'id', 'unknown')}.csv"
                )
            else:
                prompt_file = (
                    prompts_dir
                    / f"prompt_{prompt.name or getattr(prompt, 'id', 'unknown')}.json"
                )

            # Check if file already exists and should be skipped
            if should_skip_file(prompt_file, force):
                if debug:
                    debug_print(
                        f"Skipping prompt {prompt.name or prompt_id} (already exists)",
                        debug,
                    )
                else:
                    console.print(
                        f"[yellow]Skipping prompt: {prompt.name or prompt_id} (already exists)[/yellow]"
                    )
                skipped_count += 1
                continue

            # Get prompt history - use appropriate method based on prompt type
            prompt_history: List[Union[Prompt, ChatPrompt]]
            if isinstance(prompt, ChatPrompt):
                prompt_history = list(client.get_chat_prompt_history(prompt.name))
            else:
                prompt_history = list(client.get_prompt_history(prompt_id))

            # Create prompt data structure
            prompt_data = {
                "prompt": {
                    "id": getattr(prompt, "id", None),
                    "name": prompt.name,
                    "description": getattr(prompt, "description", None),
                    "created_at": (
                        created_at.isoformat()
                        if (created_at := getattr(prompt, "created_at", None))
                        else None
                    ),
                    "last_updated_at": (
                        last_updated_at.isoformat()
                        if (last_updated_at := getattr(prompt, "last_updated_at", None))
                        else None
                    ),
                },
                "current_version": {
                    "prompt": _get_prompt_content(prompt),
                    "metadata": getattr(prompt, "metadata", None),
                    "type": _get_prompt_type_string(prompt),
                    "commit": getattr(prompt, "commit", None),
                    "template_structure": _get_template_structure(prompt),
                },
                "history": [
                    {
                        "prompt": _get_prompt_content(version),
                        "metadata": getattr(version, "metadata", None),
                        "type": _get_prompt_type_string(version),
                        "commit": getattr(version, "commit", None),
                        "template_structure": _get_template_structure(version),
                    }
                    for version in prompt_history
                ],
                "downloaded_at": datetime.now().isoformat(),
            }

            # Save prompt data using the appropriate format
            if format.lower() == "csv":
                write_csv_data(prompt_data, prompt_file, prompt_to_csv_rows)
            else:
                write_json_data(prompt_data, prompt_file)

            console.print(f"[green]Exported prompt: {prompt.name or prompt_id}[/green]")
            exported_count += 1

        except Exception as e:
            if debug:
                console.print(
                    f"[yellow]Warning: Could not export prompt {prompt_id}: {e}[/yellow]"
                )
            else:
                console.print(f"[red]Error exporting prompt {prompt_id}: {e}[/red]")
            continue

    return exported_count, skipped_count


def export_related_prompts_by_name(
    client: opik.Opik,
    experiment: Any,
    output_dir: Path,
    force: bool,
    debug: bool,
    format: str = "json",
) -> int:
    """Export prompts explicitly related to the experiment from experiment metadata."""
    try:
        prompts_dir = output_dir.parent / "prompts"
        prompts_dir.mkdir(parents=True, exist_ok=True)

        # Get experiment data to access metadata
        experiment_data = experiment.get_experiment_data()
        if not experiment_data:
            debug_print("Could not get experiment data", debug)
            return 0

        # Extract prompt names from experiment metadata
        prompt_names = []
        metadata = experiment_data.metadata

        if metadata:
            # Metadata can be a dict, list, or string (JsonListStringPublic)
            # Parse if it's a string, otherwise use directly
            if isinstance(metadata, str):
                try:
                    metadata = json.loads(metadata)
                except (json.JSONDecodeError, Exception) as e:
                    if debug:
                        debug_print(f"Could not parse metadata as JSON: {e}", debug)
                    metadata = None

            # Check if metadata is a dict and has "prompts" key
            if isinstance(metadata, dict) and "prompts" in metadata:
                prompts_dict = metadata["prompts"]
                if isinstance(prompts_dict, dict):
                    # Prompts are stored as a dict with prompt names as keys
                    prompt_names = list(prompts_dict.keys())
                    if debug:
                        debug_print(
                            f"Found {len(prompt_names)} prompt(s) in experiment metadata: {prompt_names}",
                            debug,
                        )
                else:
                    if debug:
                        debug_print(
                            f"Metadata 'prompts' is not a dict, got: {type(prompts_dict)}",
                            debug,
                        )
            else:
                if debug:
                    debug_print("No 'prompts' key found in experiment metadata", debug)

        if not prompt_names:
            debug_print("No prompts found in experiment metadata", debug)
            return 0

        console.print(
            f"[blue]Exporting {len(prompt_names)} prompt(s) from experiment metadata...[/blue]"
        )

        exported_count = 0
        # Export each prompt by name from metadata
        for prompt_name in prompt_names:
            try:
                debug_print(f"Exporting prompt: {prompt_name}", debug)

                # Try to get the prompt - try ChatPrompt first, then regular Prompt
                prompt: Optional[Union[Prompt, ChatPrompt]] = None
                try:
                    prompt = client.get_chat_prompt(prompt_name)
                except Exception:
                    # Not a ChatPrompt, try regular Prompt
                    try:
                        prompt = client.get_prompt(prompt_name)
                    except Exception as e:
                        if debug:
                            console.print(
                                f"[yellow]Warning: Could not get prompt '{prompt_name}': {e}[/yellow]"
                            )
                        continue

                if not prompt:
                    if debug:
                        console.print(
                            f"[yellow]Warning: Prompt '{prompt_name}' not found[/yellow]"
                        )
                    continue

                # Get prompt history - use appropriate method based on prompt type
                prompt_history: List[Union[Prompt, ChatPrompt]]
                if isinstance(prompt, ChatPrompt):
                    prompt_history = list(client.get_chat_prompt_history(prompt.name))
                else:
                    prompt_history = list(client.get_prompt_history(prompt.name))

                # Create prompt data structure
                prompt_data = {
                    "prompt": {
                        "id": getattr(prompt, "__internal_api__prompt_id__", None),
                        "name": prompt.name,
                        "description": getattr(prompt, "description", None),
                        "created_at": getattr(prompt, "created_at", None),
                        "last_updated_at": getattr(prompt, "last_updated_at", None),
                    },
                    "current_version": {
                        "prompt": _get_prompt_content(prompt),
                        "metadata": getattr(prompt, "metadata", None),
                        "type": _get_prompt_type_string(prompt),
                        "commit": getattr(prompt, "commit", None),
                        "template_structure": _get_template_structure(prompt),
                    },
                    "history": [
                        {
                            "prompt": _get_prompt_content(version),
                            "metadata": getattr(version, "metadata", None),
                            "type": _get_prompt_type_string(version),
                            "commit": getattr(version, "commit", None),
                            "template_structure": _get_template_structure(version),
                        }
                        for version in prompt_history
                    ],
                    "downloaded_at": datetime.now().isoformat(),
                    "related_to_experiment": experiment.name or experiment.id,
                }

                # Save prompt data using the appropriate format
                # Sanitize prompt name for filename (replace / with _)
                sanitized_name = prompt.name.replace("/", "_")
                if format.lower() == "csv":
                    prompt_file = prompts_dir / f"prompts_{sanitized_name}.csv"
                else:
                    prompt_file = prompts_dir / f"prompt_{sanitized_name}.json"

                # Check if file should be skipped using the standard utility
                if should_skip_file(prompt_file, force):
                    debug_print(
                        f"Skipping prompt {prompt.name} (already exists)", debug
                    )
                    continue

                # File doesn't exist or force is set, so export it
                if format.lower() == "csv":
                    write_csv_data(prompt_data, prompt_file, prompt_to_csv_rows)
                else:
                    write_json_data(prompt_data, prompt_file)

                console.print(f"[green]Exported related prompt: {prompt.name}[/green]")
                exported_count += 1

            except Exception as e:
                if debug:
                    prompt_display_name = prompt.name if prompt else prompt_name
                    console.print(
                        f"[yellow]Warning: Could not export related prompt {prompt_display_name}: {e}[/yellow]"
                    )
                continue

        return exported_count

    except Exception as e:
        if debug:
            console.print(
                f"[yellow]Warning: Could not export related prompts: {e}[/yellow]"
            )
        return 0


@click.command(name="prompt")
@click.argument("name", type=str)
@click.option(
    "--max-results",
    type=int,
    help="Maximum number of prompts to export. Limits the total number of prompts downloaded.",
)
@click.option(
    "--path",
    "-p",
    type=click.Path(file_okay=False, dir_okay=True, writable=True),
    default="opik_exports",
    help="Directory to save exported data. Defaults to opik_exports.",
)
@click.option(
    "--force",
    is_flag=True,
    help="Re-download items even if they already exist locally.",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Enable debug output to show detailed information about the export process.",
)
@click.option(
    "--format",
    type=click.Choice(["json", "csv"], case_sensitive=False),
    default="json",
    help="Format for exporting data. Defaults to json.",
)
@click.pass_context
def export_prompt_command(
    ctx: click.Context,
    name: str,
    max_results: Optional[int],
    path: str,
    force: bool,
    debug: bool,
    format: str,
) -> None:
    """Export a prompt by exact name to workspace/prompts."""
    # Get workspace and API key from context
    workspace = ctx.obj["workspace"]
    api_key = ctx.obj.get("api_key") if ctx.obj else None
    export_prompt_by_name(
        name, workspace, path, max_results, force, debug, format, api_key
    )



import logging
from contextlib import contextmanager
from typing import Any, Generator, Optional, Dict, List

from opik import datetime_helpers
from opik.api_objects import trace, opik_client, helpers
from .. import base_track_decorator, error_info_collector

LOGGER = logging.getLogger(__name__)


@contextmanager
def start_as_current_trace(
    name: str,
    input: Optional[Dict[str, Any]] = None,
    output: Optional[Dict[str, Any]] = None,
    tags: Optional[List[str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    project_name: Optional[str] = None,
    thread_id: Optional[str] = None,
    flush: bool = False,
) -> Generator[trace.TraceData, Any, None]:
    """
    Starts a trace context manager to collect and manage tracing data during the
    execution of a code block. This function initializes a trace, allows for
    modifications within the context, and finishes the trace, sending data to the
    Opik tracing infrastructure.

    Args:
        name: The name of the trace for identification purposes.
        input: Optional input data associated with the trace.
        output: Optional output data expected or associated with the trace.
        tags: Optional list of string tags for labeling or describing the trace.
        metadata: Optional dictionary containing additional information about the
            trace.
        project_name: Optional name of the project under which the trace belongs.
        thread_id: Optional thread identifier to associate the trace with a
            specific thread.
        flush: A boolean indicating whether to flush the trace data immediately
            after finishing the trace context.

    Yields:
        Provides the initialized trace data for manipulation during execution in the context.
    """
    trace_data = trace.TraceData(
        id=helpers.generate_id(),
        start_time=datetime_helpers.local_timestamp(),
        name=name,
        input=input,
        output=output,
        metadata=metadata,
        tags=tags,
        project_name=project_name,
        thread_id=thread_id,
    )
    base_track_decorator.add_start_trace_candidate(
        trace_data=trace_data,
        opik_args_data=None,
        tracing_active=True,
    )

    try:
        yield trace_data
    except Exception as exception:
        LOGGER.error(
            "Error in user's script while executing trace context manager: %s",
            str(exception),
            exc_info=True,
        )
        trace_data.error_info = error_info_collector.collect(exception)
        trace_data.output = None
        raise
    finally:
        # save trace data at the end of the context manager
        client = opik_client.get_client_cached()
        client.trace(**trace_data.init_end_time().as_parameters)

        # Clean up trace from context
        base_track_decorator.pop_end_candidate_trace_data()

        if flush:
            client.flush()



from typing import Literal
from ..types import Event, Hint
from . import event_filter


class FilterByCount(event_filter.EventFilter):
    def __init__(self, max_count: int, level: Literal["error", "warning", "info"]):
        super().__init__()
        self._count = 0
        self._level = level
        self._max_count = max_count

    def process_event(self, event: Event, hint: Hint) -> bool:
        if not event.get("level", "") == self._level:
            return True

        if self._count >= self._max_count:
            return False

        self._count += 1
        return True



from __future__ import annotations

from typing import Any, Callable, Dict, List, Literal, Mapping, Optional, Union, Set

import enum


class PromptType(str, enum.Enum):
    MUSTACHE = "mustache"
    JINJA2 = "jinja2"


# Core multimodal/chat prompt related types
MessageContent = Union[str, List[Dict[str, Any]]]
ContentPart = Dict[str, Any]
RendererFn = Callable[[ContentPart, Dict[str, Any], PromptType], Optional[ContentPart]]
ModalityName = Literal["vision", "video"]
SupportedModalities = Mapping[ModalityName, bool]
ModalitySet = Set[ModalityName]

__all__ = [
    "PromptType",
    "MessageContent",
    "ContentPart",
    "RendererFn",
    "ModalityName",
    "SupportedModalities",
    "ModalitySet",
]



from typing import Optional, Dict, Any
from . import base_original_provider_usage


class AnthropicUsage(base_original_provider_usage.BaseOriginalProviderUsage):
    """Anthropic calls token usage data. Updated 11.03.2025"""

    input_tokens: int
    """The number of input tokens which were used."""

    output_tokens: int
    """The number of output tokens which were used."""

    cache_creation_input_tokens: Optional[int] = None
    """The number of input tokens used to create the cache entry."""

    cache_read_input_tokens: Optional[int] = None
    """The number of input tokens read from the cache."""

    def to_backend_compatible_flat_dict(self, parent_key_prefix: str) -> Dict[str, int]:
        return super().to_backend_compatible_flat_dict(parent_key_prefix)

    @classmethod
    def from_original_usage_dict(cls, usage: Dict[str, Any]) -> "AnthropicUsage":
        return cls(**usage)



from typing import Optional

import aisuite

from . import aisuite_decorator


def track_aisuite(
    aisuite_client: aisuite.Client,
    project_name: Optional[str] = None,
) -> aisuite.Client:
    """Adds Opik tracking to an AISuite client.

    Tracks calls to:
    * `aisuite_client.chat.completions.create()`,

    Can be used within other Opik-tracked functions.

    Args:
        aisuite_client: An instance of AISuite client.
        project_name: The name of the project to log data.

    Returns:
        The modified AISuite client with Opik tracking enabled.
    """

    if hasattr(aisuite_client, "opik_tracked"):
        return aisuite_client

    aisuite_client.opik_tracked = True

    decorator_factory = aisuite_decorator.AISuiteTrackDecorator()

    completions_create_decorator = decorator_factory.track(
        type="llm",
        name="chat_completion_create",
        project_name=project_name,
    )

    aisuite_client.chat.completions.create = completions_create_decorator(
        aisuite_client.chat.completions.create
    )

    return aisuite_client



import logging
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)
from typing_extensions import override

import opik.dict_utils as dict_utils
import opik.llm_usage as llm_usage
from opik.api_objects import span
from opik.decorator import arguments_helpers, base_track_decorator
from opik.types import LLMProvider

import litellm
import litellm.types.utils
import litellm.litellm_core_utils.streaming_handler

from . import stream_patchers, completion_chunks_aggregator

LOGGER = logging.getLogger(__name__)

KWARGS_KEYS_TO_LOG_AS_INPUTS: List[str] = [
    "messages",
    "functions",
    "function_call",
    "tools",
    "tool_choice",
    "response_format",
    "stop",
]
RESPONSE_KEYS_TO_LOG_AS_OUTPUT: List[str] = ["choices"]

# Sensitive parameters that should never be logged
SENSITIVE_PARAMS_TO_EXCLUDE: List[str] = [
    "api_key",
    "aws_access_key_id",
    "aws_secret_access_key",
    "azure_ad_token",
    "azure_ad_token_provider",
    "azure_key",
    "azure_password",
    "azure_username",
    "client_secret",
    "vertex_credentials",
    "vertex_project",
    "vertex_location",
    "anthropic_api_key",
    "openai_api_key",
    "cohere_api_key",
    "replicate_api_key",
    "huggingface_api_key",
    "togetherai_api_key",
    "baseten_api_key",
    "openrouter_api_key",
]

PROVIDER_MAPPING: Dict[str, LLMProvider] = {
    "openai": LLMProvider.OPENAI,
    "vertex_ai": LLMProvider.GOOGLE_VERTEXAI,
    "vertex_ai-language-models": LLMProvider.GOOGLE_VERTEXAI,
    "gemini": LLMProvider.GOOGLE_AI,
    "anthropic": LLMProvider.ANTHROPIC,
    "vertex_ai-anthropic_models": LLMProvider.ANTHROPIC_VERTEXAI,
    "bedrock": LLMProvider.BEDROCK,
    "bedrock_converse": LLMProvider.BEDROCK,
    "groq": LLMProvider.GROQ,
}


def _extract_provider_from_model(model_name: str) -> Optional[LLMProvider]:
    try:
        provider_info = litellm.get_llm_provider(model_name)
        provider_name = provider_info[1] if len(provider_info) > 1 else None
        if provider_name is None:
            return None
        return PROVIDER_MAPPING.get(provider_name, None)
    except Exception:
        return None


def _convert_response_to_dict(
    output: Union[
        litellm.types.utils.ModelResponse,
        Dict[str, Any],
    ],
) -> Dict[str, Any]:
    if hasattr(output, "model_dump"):
        return output.model_dump(mode="json")
    elif isinstance(output, dict):
        return output
    else:
        return dict(output)


def _extract_usage_from_response(
    response_dict: Dict[str, Any],
) -> Optional[llm_usage.OpikUsage]:
    usage_data = response_dict.get("usage")
    if usage_data is None:
        return None

    opik_usage = llm_usage.try_build_opik_usage_or_log_error(
        provider=LLMProvider.OPENAI,
        usage=usage_data,
        logger=LOGGER,
        error_message="Failed to log token usage from litellm call",
    )

    if opik_usage is None:
        opik_usage = llm_usage.build_opik_usage_from_unknown_provider(
            usage=usage_data,
        )

    return opik_usage


def _calculate_completion_cost(
    output: Union[
        litellm.types.utils.ModelResponse,
        Dict[str, Any],
    ],
) -> Optional[float]:
    try:
        return litellm.completion_cost(completion_response=output)
    except Exception as exception:
        LOGGER.debug(
            "Failed to calculate cost from litellm response: %s",
            str(exception),
            exc_info=True,
        )
        return None


class LiteLLMCompletionTrackDecorator(base_track_decorator.BaseTrackDecorator):
    @override
    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> arguments_helpers.StartSpanParameters:
        name = track_options.name if track_options.name is not None else func.__name__
        metadata = track_options.metadata if track_options.metadata is not None else {}

        # Filter out sensitive parameters before logging
        filtered_kwargs = {
            key: value
            for key, value in kwargs.items()
            if key not in SENSITIVE_PARAMS_TO_EXCLUDE
        }

        input_data, new_metadata = dict_utils.split_dict_by_keys(
            filtered_kwargs, keys=KWARGS_KEYS_TO_LOG_AS_INPUTS
        )
        metadata = dict_utils.deepmerge(metadata, new_metadata)
        metadata["created_from"] = "litellm"

        model_name = kwargs.get("model", "")
        provider = _extract_provider_from_model(model_name)

        return arguments_helpers.StartSpanParameters(
            name=name,
            input=input_data,
            type=track_options.type,
            tags=["litellm"],
            metadata=metadata,
            project_name=track_options.project_name,
            model=model_name,
            provider=provider,
        )

    @override
    def _end_span_inputs_preprocessor(
        self,
        output: Any,
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        assert isinstance(
            output,
            (
                litellm.types.utils.ModelResponse,
                dict,
            ),
        ), f"Expected ModelResponse or dict, got {type(output)}"

        response_dict = _convert_response_to_dict(output)
        output_data, metadata = dict_utils.split_dict_by_keys(
            response_dict, RESPONSE_KEYS_TO_LOG_AS_OUTPUT
        )

        model = response_dict.get("model")
        provider = _extract_provider_from_model(model) if model else None
        opik_usage = _extract_usage_from_response(response_dict)
        total_cost = _calculate_completion_cost(output)

        return arguments_helpers.EndSpanParameters(
            output=output_data,
            usage=opik_usage,
            metadata=metadata,
            model=model,
            provider=provider.value if provider else None,
            total_cost=total_cost,
        )

    @override
    def _streams_handler(  # type: ignore
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[
            Callable[
                [List[litellm.types.utils.ModelResponse]],
                Optional[litellm.types.utils.ModelResponse],
            ]
        ],
    ) -> Optional[litellm.litellm_core_utils.streaming_handler.CustomStreamWrapper]:
        assert generations_aggregator is not None, (
            "LiteLLM decorator will always get aggregator function as input"
        )

        is_litellm_stream = isinstance(
            output, litellm.litellm_core_utils.streaming_handler.CustomStreamWrapper
        )
        if not is_litellm_stream:
            return None

        span_to_end, trace_to_end = base_track_decorator.pop_end_candidates()

        return stream_patchers.patch_stream(
            stream=output,
            span_to_end=span_to_end,
            trace_to_end=trace_to_end,
            generations_aggregator=completion_chunks_aggregator.aggregate,
            finally_callback=self._after_call,
        )



from typing import Any, Dict, List, Mapping, MutableMapping, Optional, Tuple, Union

from .. import types as prompt_types


class ChatContentRendererRegistry:
    """
    Registry that knows how to render structured content parts.
    """

    def __init__(self) -> None:
        self._part_renderers: MutableMapping[str, prompt_types.RendererFn] = {}
        self._part_modalities: MutableMapping[
            str, Optional[prompt_types.ModalityName]
        ] = {}
        self._modality_placeholders: MutableMapping[
            prompt_types.ModalityName, Tuple[str, str]
        ] = {
            "vision": ("<<<image>>>", "<<</image>>>"),
            "video": ("<<<video>>>", "<<</video>>>"),
        }
        self._default_placeholder: Tuple[str, str] = ("<<<media>>>", "<<</media>>>")
        self._placeholder_value_limit = 500

    def register_part_renderer(
        self,
        part_type: str,
        renderer: prompt_types.RendererFn,
        *,
        modality: Optional[prompt_types.ModalityName] = None,
        placeholder: Optional[Tuple[str, str]] = None,
    ) -> None:
        """
        Register or override the renderer responsible for a structured content part.
        """
        normalized_part = part_type.lower()
        self._part_renderers[normalized_part] = renderer
        self._part_modalities[normalized_part] = modality

        if modality and placeholder:
            self._modality_placeholders[modality] = placeholder

    def render_content(
        self,
        content: prompt_types.MessageContent,
        variables: Dict[str, Any],
        template_type: prompt_types.PromptType,
        *,
        supported_modalities: Optional[prompt_types.SupportedModalities] = None,
    ) -> prompt_types.MessageContent:
        if supported_modalities is None:
            modality_flags: Dict[str, bool] = {}
        else:
            modality_flags = {
                modality: bool(is_supported)
                for modality, is_supported in supported_modalities.items()
            }

        if isinstance(content, str):
            # Resolved in ChatPromptTemplate
            return content

        if not isinstance(content, list):
            return str(content)

        if not content:
            return []

        rendered_parts = self._render_structured_content(
            content=content, variables=variables, template_type=template_type
        )

        if not rendered_parts:
            return []

        if self._should_flatten(rendered_parts, modality_flags):
            return self._flatten_parts_to_text(rendered_parts, modality_flags)

        return rendered_parts

    def normalize_template_type(
        self, template_type: Union[str, prompt_types.PromptType]
    ) -> prompt_types.PromptType:
        if isinstance(template_type, prompt_types.PromptType):
            return template_type
        try:
            return prompt_types.PromptType(template_type)
        except ValueError:
            return prompt_types.PromptType.MUSTACHE

    def infer_modalities(
        self, content: prompt_types.MessageContent
    ) -> set[prompt_types.ModalityName]:
        if not isinstance(content, list):
            return set()
        modalities: set[prompt_types.ModalityName] = set()
        for part in content:
            if not isinstance(part, dict):
                continue
            part_type = part.get("type", "").lower()
            modality = self._part_modalities.get(part_type)
            if modality:
                modalities.add(modality)
        return modalities

    def _render_structured_content(
        self,
        content: List[Any],
        variables: Dict[str, Any],
        template_type: prompt_types.PromptType,
    ) -> List[prompt_types.ContentPart]:
        rendered_parts: List[prompt_types.ContentPart] = []
        for part in content:
            if not isinstance(part, dict):
                continue
            part_type = part.get("type", "").lower()
            renderer = self._part_renderers.get(part_type)
            if renderer is None:
                continue
            rendered_part = renderer(part, variables, template_type)
            if rendered_part is None:
                continue
            rendered_parts.append(rendered_part)
        return rendered_parts

    def _should_flatten(
        self, parts: List[prompt_types.ContentPart], modality_flags: Mapping[str, bool]
    ) -> bool:
        for part in parts:
            modality = self._part_modalities.get(part.get("type", "").lower())
            if modality and not modality_flags.get(modality, False):
                return True
        return False

    def _flatten_parts_to_text(
        self, parts: List[prompt_types.ContentPart], modality_flags: Mapping[str, bool]
    ) -> str:
        segments: List[str] = []
        for part in parts:
            part_type = part.get("type", "").lower()
            if part_type == "text":
                text = part.get("text", "")
                if text:
                    segments.append(text)
                continue
            modality = self._part_modalities.get(part_type)
            if modality and modality_flags.get(modality, False):
                segments.append(str(part))
                continue
            if modality and modality in self._modality_placeholders:
                prefix, suffix = self._modality_placeholders[modality]
            else:
                prefix, suffix = self._default_placeholder
            placeholder_value = self._extract_placeholder_value(part)
            if placeholder_value:
                truncated = self._truncate_placeholder_value(placeholder_value)
                segments.append(f"{prefix}{truncated}{suffix}")
        return "\n\n".join(segment for segment in segments if segment)

    @staticmethod
    def _extract_placeholder_value(part: prompt_types.ContentPart) -> str:
        part_type = part.get("type", "").lower()
        if part_type == "image_url":
            image_dict = part.get("image_url", {})
            if isinstance(image_dict, dict):
                return str(image_dict.get("url", "")).strip()
        if part_type == "video_url":
            video_dict = part.get("video_url", {})
            if isinstance(video_dict, dict):
                return str(video_dict.get("url", "")).strip()
        return str(part)

    def _truncate_placeholder_value(self, value: str) -> str:
        if len(value) <= self._placeholder_value_limit:
            return value
        if self._placeholder_value_limit <= 3:
            return value[: self._placeholder_value_limit]
        return value[: self._placeholder_value_limit - 3] + "..."


DEFAULT_CHAT_RENDERER_REGISTRY = ChatContentRendererRegistry()


def register_default_chat_part_renderer(
    part_type: str,
    renderer: prompt_types.RendererFn,
    *,
    modality: Optional[prompt_types.ModalityName] = None,
    placeholder: Optional[Tuple[str, str]] = None,
) -> None:
    DEFAULT_CHAT_RENDERER_REGISTRY.register_part_renderer(
        part_type,
        renderer,
        modality=modality,
        placeholder=placeholder,
    )


__all__ = [
    "ChatContentRendererRegistry",
    "DEFAULT_CHAT_RENDERER_REGISTRY",
    "register_default_chat_part_renderer",
]



import base64
import binascii
import logging
import tempfile
from typing import Any, Optional, Literal

from . import attachment, decoder, decoder_helpers

LOGGER = logging.getLogger(__name__)


class Base64AttachmentDecoder(decoder.AttachmentDecoder):
    """Decodes base64 encoded attachment data.

    This decoder decodes base64 strings, detects MIME types from content, and creates Attachment objects.
    """

    def decode(
        self,
        raw_data: str,
        context: Literal["input", "output", "metadata"] = "input",
        **kwargs: Any,
    ) -> Optional[attachment.Attachment]:
        """Decode base64 encoded data into an Attachment object.

        Args:
            raw_data: Base64 encoded string data
            context: Context string for filename generation.

        Returns:
            Attachment object with decoded data, or None if decoding fails or type is not recognizable
        """
        if not isinstance(raw_data, str):
            LOGGER.warning("Attachment data is not a string, skipping.")
            return None

        try:
            # Decode base64 string to bytes
            decoded_bytes = base64.b64decode(raw_data, validate=True)

            # Detect MIME type from content
            mime_type = decoder_helpers.detect_mime_type(decoded_bytes)

            # Skip if not a recognizable file type
            if not mime_type or mime_type in ("application/octet-stream", "text/plain"):
                LOGGER.debug("Attachment type is not recognized, skipping.")
                return None

            # Get file extension from the MIME type
            extension = decoder_helpers.get_file_extension(mime_type)

            # Generate filename
            file_name = decoder_helpers.create_attachment_filename(
                context, extension=extension
            )

            # Save decoded bytes to a temporary file
            temp_file = tempfile.NamedTemporaryFile(
                mode="wb", delete=False, suffix=extension
            )
            temp_file.write(decoded_bytes)
            temp_file.flush()
            temp_file.close()

            # Return Attachment object with a file path
            return attachment.Attachment(
                data=temp_file.name, file_name=file_name, content_type=mime_type
            )

        except (ValueError, binascii.Error) as e:
            LOGGER.debug(
                "Failed to decode attachment data, reason: invalid base64. Reason: %s",
                e,
                exc_info=True,
            )
            # Not valid base64, return None
            return None
        except Exception as ex:
            LOGGER.warning(
                "Failed to decode attachment data, reason: %s", ex, exc_info=True
            )
            # Unexpected error, return None to avoid crashing the pipeline
            return None



from typing import Optional, Dict, Any

import pydantic


LANGCHAIN_USAGE_KEYS = {"input_tokens", "output_tokens", "total_tokens"}


class InputTokenDetails(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")

    cache_read: Optional[int] = None
    """Number of tokens read from the cache."""

    cache_creation: Optional[int] = None
    """Number of tokens created in the cache."""

    audio: Optional[int] = None
    """Number of tokens in the audio prompt."""


class OutputTokenDetails(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="allow")

    reasoning: Optional[int] = None
    """Number of tokens generated by the model for reasoning."""

    audio: Optional[int] = None
    """Number of tokens in the audio response."""


class LangChainUsage(pydantic.BaseModel):
    # """
    # LangChain Usage class, based on, but not inherited from UsageMetadata class in langchain_code.messages.ai.UsageMetadata
    # https://python.langchain.com/api_reference/core/messages/langchain_core.messages.ai.UsageMetadata.html#usagemetadata
    # """
    model_config = pydantic.ConfigDict(extra="allow")

    input_tokens: int
    """Number of tokens in the prompt."""

    output_tokens: int
    """Number of tokens in the response(s)."""

    total_tokens: int
    """Total token count for prompt, response candidates, and reasoning."""

    input_token_details: Optional[InputTokenDetails]
    """Breakdown of tokens used in the prompt."""

    output_token_details: Optional[OutputTokenDetails]
    """Breakdown of tokens used in a response."""

    @classmethod
    def from_original_usage_dict(cls, usage: Dict[str, Any]) -> "LangChainUsage":
        usage_dict = {**usage}
        input_token_details_raw = usage_dict.pop("input_token_details", None)
        output_token_details_raw = usage_dict.pop("output_token_details", None)

        input_token_details = (
            InputTokenDetails(**input_token_details_raw)
            if isinstance(input_token_details_raw, dict)
            else None
        )

        output_token_details = (
            OutputTokenDetails(**output_token_details_raw)
            if isinstance(output_token_details_raw, dict)
            else None
        )

        return cls(
            **usage_dict,
            input_token_details=input_token_details,
            output_token_details=output_token_details,
        )

    def map_to_google_gemini_usage(self) -> Dict[str, Any]:
        google_usage: Dict[str, Any] = {
            "prompt_token_count": self.input_tokens,
            "candidates_token_count": self.output_tokens,
            "total_token_count": self.total_tokens,
        }
        if self.input_token_details is not None:
            google_usage["cached_content_token_count"] = (
                self.input_token_details.cache_read
            )

        if self.output_token_details is not None:
            google_usage["thoughts_token_count"] = self.output_token_details.reasoning

        return google_usage

    def map_to_anthropic_usage(self) -> Dict[str, Any]:
        anthropic_usage: Dict[str, Any] = {
            "input_tokens": self.input_tokens,
            "output_tokens": self.output_tokens,
            "total_tokens": self.total_tokens,
        }
        if self.input_token_details is None:
            return anthropic_usage

        if self.input_token_details.cache_creation is not None:
            anthropic_usage["cache_creation_input_tokens"] = (
                self.input_token_details.cache_creation
            )
        if self.input_token_details.cache_read is not None:
            anthropic_usage["cache_read_input_tokens"] = (
                self.input_token_details.cache_read
            )

        return anthropic_usage

    def map_to_bedrock_usage(self) -> Dict[str, Any]:
        bedrock_usage: Dict[str, Any] = {
            "inputTokens": self.input_tokens,
            "outputTokens": self.output_tokens,
        }

        if self.input_token_details is not None:
            bedrock_usage["cacheReadInputTokens"] = self.input_token_details.cache_read
            bedrock_usage["cacheWriteInputTokens"] = (
                self.input_token_details.cache_creation
            )

        return bedrock_usage

    def map_to_openai_completions_usage(self) -> Dict[str, Any]:
        openai_usage: Dict[str, Any] = {
            "prompt_tokens": self.input_tokens,
            "completion_tokens": self.output_tokens,
            "total_tokens": self.total_tokens,
        }

        if self.input_token_details is not None:
            openai_usage["prompt_tokens_details"] = {
                "cached_tokens": self.input_token_details.cache_read,
                "audio_tokens": self.input_token_details.audio,
            }

        if self.output_token_details is not None:
            openai_usage["completion_tokens_details"] = {
                "audio_tokens": self.output_token_details.audio,
                "reasoning_tokens": self.output_token_details.reasoning,
            }

        return openai_usage

    def map_to_groq_completions_usage(self) -> Dict[str, Any]:
        groq_usage: Dict[str, Any] = {
            "prompt_tokens": self.input_tokens,
            "completion_tokens": self.output_tokens,
            "total_tokens": self.total_tokens,
        }

        return groq_usage


def is_langchain_usage(usage_dict: Dict[str, Any]) -> bool:
    if usage_dict is None:
        return False

    return all(key in usage_dict for key in LANGCHAIN_USAGE_KEYS)



import abc
import datetime
import logging
import functools
import sys
from typing import (
    Optional,
    Any,
    List,
    Dict,
    Sequence,
    Set,
    TYPE_CHECKING,
    Iterator,
)

from opik.api_objects import rest_helpers
from opik.rest_api import client as rest_api_client
from opik.rest_api.types import (
    dataset_item_write as rest_dataset_item,
    dataset_version_public,
)
from opik.message_processing.batching import sequence_splitter
from opik import id_helpers
import opik.exceptions as exceptions
import opik.config as config
from .. import constants
from . import dataset_item, converters, rest_operations

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

if TYPE_CHECKING:
    import pandas as pd

LOGGER = logging.getLogger(__name__)


class DatasetExportOperations(abc.ABC):
    """
    Abstract base class providing export operations for dataset items.

    This class defines the common interface for exporting dataset items,
    shared by both Dataset (current state) and DatasetVersion (specific version).
    """

    @abc.abstractmethod
    def __internal_api__stream_items_as_dataclasses__(
        self,
        nb_samples: Optional[int] = None,
        batch_size: Optional[int] = None,
        dataset_item_ids: Optional[List[str]] = None,
        filter_string: Optional[str] = None,
    ) -> Iterator[dataset_item.DatasetItem]:
        """
        Stream dataset items as DatasetItem objects.

        Args:
            nb_samples: Maximum number of items to retrieve.
            batch_size: Maximum number of items to fetch per batch.
            dataset_item_ids: Optional list of specific item IDs to retrieve.
            filter_string: Optional OQL filter string to filter dataset items.

        Yields:
            DatasetItem objects one at a time.
        """
        raise NotImplementedError

    def to_pandas(self) -> "pd.DataFrame":
        """
        Convert the dataset items to a pandas DataFrame.

        Requires the `pandas` library to be installed.

        Returns:
            A pandas DataFrame containing all items.
        """
        dataset_items = list(self.__internal_api__stream_items_as_dataclasses__())
        return converters.to_pandas(dataset_items, keys_mapping={})

    def to_json(self) -> str:
        """
        Convert the dataset items to a JSON string.

        Returns:
            A JSON string representation of all items.
        """
        dataset_items = list(self.__internal_api__stream_items_as_dataclasses__())
        return converters.to_json(dataset_items, keys_mapping={})

    def get_items(
        self,
        nb_samples: Optional[int] = None,
        filter_string: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        Retrieve dataset items as a list of dictionaries.

        Args:
            nb_samples: Maximum number of items to retrieve. If not set, all items are returned.
            filter_string: Optional OQL filter string to filter dataset items.
                Supports filtering by tags, data fields, metadata, etc.

                Supported columns include:
                - `id`, `source`, `trace_id`, `span_id`: String fields
                - `data`: Dictionary field (use dot notation, e.g., "data.category")
                - `tags`: List field (use "contains" operator)
                - `created_at`, `last_updated_at`: DateTime fields (ISO 8601 format)
                - `created_by`, `last_updated_by`: String fields

                Examples:
                - `tags contains "failed"` - Items with 'failed' tag
                - `data.category = "test"` - Items with specific data field value
                - `created_at >= "2024-01-01T00:00:00Z"` - Items created after date

        Returns:
            A list of dictionaries representing the dataset items.
        """
        dataset_items_as_dicts = [
            {"id": item.id, **item.get_content()}
            for item in self.__internal_api__stream_items_as_dataclasses__(
                nb_samples=nb_samples, filter_string=filter_string
            )
        ]
        return dataset_items_as_dicts

    @abc.abstractmethod
    def get_version_info(
        self,
    ) -> Optional[dataset_version_public.DatasetVersionPublic]:
        """
        Get version information for experiment association.

        Returns:
            DatasetVersionPublic containing version metadata (id, version_name, etc.).
            For Dataset, returns info about the current/latest version, or None if no version exists.
            For DatasetVersion, returns info about this specific version.
        """
        raise NotImplementedError


class DatasetVersion(DatasetExportOperations):
    """
    A read-only view of a specific dataset version.

    This class provides access to dataset items at a specific version point in time.
    It supports reading version metadata and retrieving items, but does not allow
    mutations to the dataset.

    This object should not be created directly. Use :meth:`Dataset.get_dataset_version`
    to obtain an instance.
    """

    def __init__(
        self,
        dataset_name: str,
        dataset_id: str,
        rest_client: rest_api_client.OpikApi,
        version_info: dataset_version_public.DatasetVersionPublic,
    ) -> None:
        self._dataset_name = dataset_name
        self._dataset_id = dataset_id
        self._rest_client = rest_client
        self._version_info = version_info

    @property
    def dataset_name(self) -> str:
        """The name of the dataset this version belongs to."""
        return self._dataset_name

    @property
    def name(self) -> str:
        """The name of the dataset this version belongs to (alias for dataset_name)."""
        return self._dataset_name

    @property
    def dataset_id(self) -> str:
        """The unique identifier of the dataset this version belongs to."""
        return self._dataset_id

    @property
    def id(self) -> str:
        """The unique identifier of the dataset this version belongs to (alias for dataset_id)."""
        return self._dataset_id

    @property
    def version_id(self) -> Optional[str]:
        """The unique identifier of this specific version."""
        return self._version_info.id

    @property
    def dataset_items_count(self) -> Optional[int]:
        """Total number of items in this version (alias for items_total)."""
        return self._version_info.items_total

    @property
    def version_hash(self) -> Optional[str]:
        """The unique hash identifier of this version."""
        return self._version_info.version_hash

    @property
    def version_name(self) -> Optional[str]:
        """The sequential version name (e.g., 'v1', 'v2')."""
        return self._version_info.version_name

    @property
    def tags(self) -> Optional[List[str]]:
        """Tags associated with this version."""
        return self._version_info.tags

    @property
    def is_latest(self) -> Optional[bool]:
        """Whether this is the latest version of the dataset."""
        return self._version_info.is_latest

    @property
    def items_total(self) -> Optional[int]:
        """Total number of items in this version."""
        return self._version_info.items_total

    @property
    def items_added(self) -> Optional[int]:
        """Number of items added since the previous version."""
        return self._version_info.items_added

    @property
    def items_modified(self) -> Optional[int]:
        """Number of items modified since the previous version."""
        return self._version_info.items_modified

    @property
    def items_deleted(self) -> Optional[int]:
        """Number of items deleted since the previous version."""
        return self._version_info.items_deleted

    @property
    def change_description(self) -> Optional[str]:
        """Description of changes in this version."""
        return self._version_info.change_description

    @property
    def created_at(self) -> Optional[datetime.datetime]:
        """Timestamp when this version was created."""
        return self._version_info.created_at

    @property
    def created_by(self) -> Optional[str]:
        """User who created this version."""
        return self._version_info.created_by

    @override
    def __internal_api__stream_items_as_dataclasses__(
        self,
        nb_samples: Optional[int] = None,
        batch_size: Optional[int] = None,
        dataset_item_ids: Optional[List[str]] = None,
        filter_string: Optional[str] = None,
    ) -> Iterator[dataset_item.DatasetItem]:
        return rest_operations.stream_dataset_items(
            rest_client=self._rest_client,
            dataset_name=self._dataset_name,
            nb_samples=nb_samples,
            batch_size=batch_size,
            dataset_item_ids=dataset_item_ids,
            filter_string=filter_string,
            dataset_version=self._version_info.version_hash,
        )

    @override
    def get_version_info(
        self,
    ) -> Optional[dataset_version_public.DatasetVersionPublic]:
        """
        Get version information for this specific dataset version.

        Returns:
            DatasetVersionPublic containing this version's metadata.
        """
        return self._version_info


class Dataset(DatasetExportOperations):
    def __init__(
        self,
        name: str,
        description: Optional[str],
        rest_client: rest_api_client.OpikApi,
        dataset_items_count: Optional[int] = None,
    ) -> None:
        """
        A Dataset object. This object should not be created directly, instead use :meth:`opik.Opik.create_dataset` or :meth:`opik.Opik.get_dataset`.
        """
        self._name = name
        self._description = description
        self._rest_client = rest_client
        self._dataset_items_count = dataset_items_count

        self._id_to_hash: Dict[str, str] = {}
        self._hashes: Set[str] = set()

    @functools.cached_property
    def id(self) -> str:
        """The id of the dataset"""
        return self._rest_client.datasets.get_dataset_by_identifier(
            dataset_name=self._name
        ).id

    @property
    def name(self) -> str:
        """The name of the dataset."""
        return self._name

    @property
    def description(self) -> Optional[str]:
        """The description of the dataset."""
        return self._description

    @property
    def dataset_items_count(self) -> Optional[int]:
        """
        The total number of items in the dataset.

        If the count is not cached locally, it will be fetched from the backend.
        """
        if self._dataset_items_count is None:
            dataset_info = self._rest_client.datasets.get_dataset_by_identifier(
                dataset_name=self._name
            )
            self._dataset_items_count = dataset_info.dataset_items_count
        return self._dataset_items_count

    def get_current_version_name(self) -> Optional[str]:
        """
        Get the current version name of the dataset.

        The version name is fetched from the backend and reflects the latest
        committed version after any mutation operations (insert, update, delete).

        Returns:
            The current version name (e.g., 'v1', 'v2'), or None if no version exists.
        """
        version_info = self.get_version_info()
        return version_info.version_name if version_info else None

    @override
    def get_version_info(
        self,
    ) -> Optional[dataset_version_public.DatasetVersionPublic]:
        """
        Get version information for the current (latest) dataset version.

        Returns:
            DatasetVersionPublic containing the current version's metadata,
            or None if no version exists yet.
        """
        versions_response = self._rest_client.datasets.list_dataset_versions(
            id=self.id,
            page=1,
            size=1,
        )
        if not versions_response.content:
            return None
        return versions_response.content[0]

    def _insert_batch_with_retry(
        self,
        batch: List[rest_dataset_item.DatasetItemWrite],
        batch_group_id: str,
    ) -> None:
        """Insert a batch of dataset items with automatic retry on rate limit errors.

        Args:
            batch: List of dataset items to insert.
            batch_group_id: UUIDv7 identifier that groups all batches from a single
                user operation together. All batches sent as part of one insert/update
                call share the same batch_group_id.
        """
        rest_helpers.ensure_rest_api_call_respecting_rate_limit(
            lambda: self._rest_client.datasets.create_or_update_dataset_items(
                dataset_name=self._name, items=batch, batch_group_id=batch_group_id
            )
        )
        LOGGER.debug("Successfully sent dataset items batch of size %d", len(batch))

    def __internal_api__insert_items_as_dataclasses__(
        self, items: List[dataset_item.DatasetItem]
    ) -> None:
        # Remove duplicates if they already exist
        deduplicated_items: List[dataset_item.DatasetItem] = []
        for item in items:
            item_hash = item.content_hash()

            if item_hash in self._hashes:
                LOGGER.debug(
                    "Duplicate item found with hash: %s - ignored the event",
                    item_hash,
                )
                continue

            deduplicated_items.append(item)
            self._hashes.add(item_hash)
            self._id_to_hash[item.id] = item_hash

        rest_items = [
            rest_dataset_item.DatasetItemWrite(
                id=item.id,  # type: ignore
                trace_id=item.trace_id,  # type: ignore
                span_id=item.span_id,  # type: ignore
                source=item.source,  # type: ignore
                data=item.get_content(),
            )
            for item in deduplicated_items
        ]

        batches = sequence_splitter.split_into_batches(
            rest_items,
            max_payload_size_MB=config.MAX_BATCH_SIZE_MB,
            max_length=constants.DATASET_ITEMS_MAX_BATCH_SIZE,
        )

        batch_group_id = id_helpers.generate_id()

        for batch in batches:
            LOGGER.debug("Sending dataset items batch of size %d", len(batch))
            self._insert_batch_with_retry(batch, batch_group_id=batch_group_id)

    def insert(self, items: Sequence[Dict[str, Any]]) -> None:
        """
        Insert new items into the dataset. A new dataset version will be created.

        Args:
            items: List of dicts (which will be converted to dataset items)
                to add to the dataset.
        """
        dataset_items: List[dataset_item.DatasetItem] = [  # type: ignore
            (dataset_item.DatasetItem(**item) if isinstance(item, dict) else item)
            for item in items
        ]
        self.__internal_api__insert_items_as_dataclasses__(dataset_items)

        # Invalidate the cached count so it will be fetched from backend on next access
        self._dataset_items_count = None

    def __internal_api__sync_hashes__(self) -> None:
        """Updates all the hashes in the dataset"""
        LOGGER.debug("Start hash sync in dataset")

        self._id_to_hash = {}
        self._hashes = set()

        for item in self.__internal_api__stream_items_as_dataclasses__():
            item_hash = item.content_hash()
            self._id_to_hash[item.id] = item_hash  # type: ignore
            self._hashes.add(item_hash)

        LOGGER.debug("Finish hash sync in dataset")

    def update(self, items: List[Dict[str, Any]]) -> None:
        """
        Update existing items in the dataset.

        Args:
            items: List of DatasetItem objects to update in the dataset. You need to provide the full item object as it will override what has been supplied previously.

        Raises:
            DatasetItemUpdateOperationRequiresItemId: If any item in the list is missing an id.
        """
        for item in items:
            if "id" not in item:
                raise exceptions.DatasetItemUpdateOperationRequiresItemId(
                    "Missing id for dataset item to update: %s", item
                )

        self.insert(items)

    def _delete_batch_with_retry(
        self,
        batch: List[str],
        batch_group_id: str,
    ) -> None:
        """Delete a batch of dataset items with automatic retry on rate limit errors.

        Args:
            batch: List of item IDs to delete.
            batch_group_id: UUIDv7 identifier that groups all batches from a single
                user operation together. All batches sent as part of one delete
                call share the same batch_group_id.
        """
        rest_helpers.ensure_rest_api_call_respecting_rate_limit(
            lambda: self._rest_client.datasets.delete_dataset_items(
                item_ids=batch, batch_group_id=batch_group_id
            )
        )
        LOGGER.debug("Successfully deleted dataset items batch of size %d", len(batch))

    def delete(self, items_ids: List[str]) -> None:
        """
        Delete items from the dataset. A new dataset version will be created.

        Args:
            items_ids: List of item ids to delete.
        """
        batches = sequence_splitter.split_into_batches(
            items_ids, max_length=constants.DATASET_ITEMS_MAX_BATCH_SIZE
        )

        batch_group_id = id_helpers.generate_id()

        for batch in batches:
            LOGGER.debug("Deleting dataset items batch: %s", batch)
            self._delete_batch_with_retry(batch, batch_group_id=batch_group_id)

            for item_id in batch:
                if item_id in self._id_to_hash:
                    hash = self._id_to_hash[item_id]
                    self._hashes.discard(hash)
                    del self._id_to_hash[item_id]

        # Invalidate the cached count so it will be fetched from backend on next access
        self._dataset_items_count = None

    def clear(self) -> None:
        """
        Delete all items from the given dataset. A new dataset version will be created.
        """
        item_ids = [
            item.id
            for item in self.__internal_api__stream_items_as_dataclasses__()
            if item.id is not None
        ]

        self.delete(item_ids)

    @override
    def __internal_api__stream_items_as_dataclasses__(
        self,
        nb_samples: Optional[int] = None,
        batch_size: Optional[int] = None,
        dataset_item_ids: Optional[List[str]] = None,
        filter_string: Optional[str] = None,
    ) -> Iterator[dataset_item.DatasetItem]:
        """
        Stream dataset items as a generator instead of loading all at once.

        This method yields dataset items one at a time, enabling evaluation to start
        processing items before the entire dataset is downloaded. This is particularly
        useful for large datasets with heavy payloads (images, videos, audio).

        Args:
            nb_samples: Maximum number of items to retrieve. If None, all items are streamed.
            batch_size: Maximum number of items to fetch per batch from the backend.
                        If None, uses the default value from constants.DATASET_STREAM_BATCH_SIZE.
            dataset_item_ids: Optional list of specific item IDs to retrieve. If provided,
                            only items with matching IDs will be yielded.
            filter_string: Optional OQL filter string to filter dataset items.

        Yields:
            DatasetItem objects one at a time
        """
        return rest_operations.stream_dataset_items(
            rest_client=self._rest_client,
            dataset_name=self._name,
            nb_samples=nb_samples,
            batch_size=batch_size,
            dataset_item_ids=dataset_item_ids,
            filter_string=filter_string,
            dataset_version=None,
        )

    def insert_from_json(
        self,
        json_array: str,
        keys_mapping: Optional[Dict[str, str]] = None,
        ignore_keys: Optional[List[str]] = None,
    ) -> None:
        """
        Args:
            json_array: json string of format: "[{...}, {...}, {...}]" where every dictionary
                is to be transformed into dataset item
            keys_mapping: dictionary that maps json keys to item fields names
                Example: {'Expected output': 'expected_output'}
            ignore_keys: if your json dicts contain keys that are not needed for DatasetItem
                construction - pass them as ignore_keys argument
        """
        keys_mapping = {} if keys_mapping is None else keys_mapping
        ignore_keys = [] if ignore_keys is None else ignore_keys

        new_items = converters.from_json(
            json_array, keys_mapping=keys_mapping, ignore_keys=ignore_keys
        )

        self.insert(new_items)

    def read_jsonl_from_file(
        self,
        file_path: str,
        keys_mapping: Optional[Dict[str, str]] = None,
        ignore_keys: Optional[List[str]] = None,
    ) -> None:
        """
        Read JSONL from a file and insert it into the dataset.

        Args:
            file_path: Path to the JSONL file
            keys_mapping: dictionary that maps json keys to item fields names
                Example: {'Expected output': 'expected_output'}
            ignore_keys: if your json dicts contain keys that are not needed for DatasetItem
                construction - pass them as ignore_keys argument
        """
        keys_mapping = {} if keys_mapping is None else keys_mapping
        ignore_keys = [] if ignore_keys is None else ignore_keys
        new_items = converters.from_jsonl_file(file_path, keys_mapping, ignore_keys)
        self.insert(new_items)

    def insert_from_pandas(
        self,
        dataframe: "pd.DataFrame",
        keys_mapping: Optional[Dict[str, str]] = None,
        ignore_keys: Optional[List[str]] = None,
    ) -> None:
        """
        Requires: `pandas` library to be installed.

        Args:
            dataframe: pandas dataframe
            keys_mapping: Dictionary that maps dataframe column names to dataset item field names.
                Example: {'Expected output': 'expected_output'}
            ignore_keys: if your dataframe contains columns that are not needed for DatasetItem
                construction - pass them as ignore_keys argument
        """
        keys_mapping = {} if keys_mapping is None else keys_mapping
        ignore_keys = [] if ignore_keys is None else ignore_keys

        new_items = converters.from_pandas(dataframe, keys_mapping, ignore_keys)

        self.insert(new_items)

    def get_version_view(self, version_name: str) -> DatasetVersion:
        """
        Get a read-only view of a specific dataset version.

        The returned DatasetVersion object allows reading version metadata and
        retrieving items via :meth:`DatasetVersion.get_items`, but does not support
        mutations.

        Args:
            version_name: The version name (e.g., 'v1', 'v2').

        Returns:
            A read-only DatasetVersion object for accessing the specified version.

        Raises:
            opik.exceptions.DatasetVersionNotFound: If the specified version does not exist.

        Example:
            >>> dataset = client.get_dataset("my_dataset")
            >>> version = dataset.get_version_view("v1")
            >>> items = version.get_items()
        """
        version_info = rest_operations.find_version_by_name(
            rest_client=self._rest_client,
            dataset_id=self.id,
            version_name=version_name,
        )

        if version_info is None:
            raise exceptions.DatasetVersionNotFound(
                f"Dataset version '{version_name}' not found in dataset '{self._name}'"
            )

        return DatasetVersion(
            dataset_name=self._name,
            dataset_id=self.id,
            rest_client=self._rest_client,
            version_info=version_info,
        )



import functools

import httpx
import tenacity

from opik import hooks

KEEPALIVE_EXPIRY_SECONDS = 10
CONNECT_TIMEOUT_SECONDS = 20
READ_TIMEOUT_SECONDS = 100
WRITE_TIMEOUT_SECONDS = 100
POOL_TIMEOUT_SECONDS = 20

RETRYABLE_STATUS_CODES = [500]


@functools.lru_cache
def get_cached() -> httpx.Client:
    return get()


def get() -> httpx.Client:
    limits = httpx.Limits(keepalive_expiry=KEEPALIVE_EXPIRY_SECONDS)

    timeout = httpx.Timeout(
        connect=CONNECT_TIMEOUT_SECONDS,
        read=READ_TIMEOUT_SECONDS,
        write=WRITE_TIMEOUT_SECONDS,
        pool=POOL_TIMEOUT_SECONDS,
    )

    # build HTTPX client arguments
    kwargs = {
        "limits": limits,
        "timeout": timeout,
    }
    kwargs = hooks.httpx_client_hook.build_init_arguments(kwargs)

    client = httpx.Client(**kwargs)

    hooks.httpx_client_hook.apply_httpx_client_hooks(client)

    return client


def _allowed_to_retry(exception: Exception) -> bool:
    if isinstance(
        exception,
        (
            httpx.ConnectError,
            httpx.TimeoutException,
        ),
    ):
        return True

    if isinstance(exception, httpx.HTTPStatusError):
        if exception.response.status_code in RETRYABLE_STATUS_CODES:
            return True

    return False


s3_retry = tenacity.retry(
    stop=tenacity.stop_after_attempt(3),
    wait=tenacity.wait_exponential(multiplier=5, min=10, max=45),
    retry=tenacity.retry_if_exception(_allowed_to_retry),
)
